@@ -1,4 +1,4 @@
-/*      $Id: focus.c 23551 2006-10-29 22:59:52Z olivier $
+/*      $Id: focus.c 29542 2009-02-23 10:06:52Z olivier $
 
         This program is free software; you can redistribute it and/or modify
         it under the terms of the GNU General Public License as published by
@@ -12,22 +12,25 @@
 
         You should have received a copy of the GNU General Public License
         along with this program; if not, write to the Free Software
-        Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+        Foundation, Inc., Inc., 51 Franklin Street, Fifth Floor, Boston,
+        MA 02110-1301, USA.
+
 
         oroborus - (c) 2001 Ken Lynch
-        xfwm4    - (c) 2002-2006 Olivier Fourdan
+        xfwm4    - (c) 2002-2009 Olivier Fourdan
 
  */
 
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+#include "config.h"
 #endif
 
+#include <X11/X.h>
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/Xatom.h>
 #include <X11/extensions/shape.h>
-#include <sys/time.h>
+
 #include <glib.h>
 #include <gdk/gdk.h>
 #include <gdk/gdkx.h>
@@ -54,10 +57,28 @@ struct _ClientPair
 
 static Client *client_focus  = NULL;
 static Client *pending_focus = NULL;
+static Client *user_focus    = NULL;
 static Client *last_ungrab   = NULL;
+static Client *delayed_focus = NULL;
+static guint focus_timeout   = 0;
+
+#if 0
+static void
+clientDumpList (ScreenInfo *screen_info)
+{
+    int i;
+    Client *c;
+
+    g_print ("Dumping client list\n");
+    for (c = screen_info->clients, i = 0; (c) && (i < screen_info->client_count); c = c->next, i++)
+    {
+        g_print ("    [%i] 0x%lx - %s\n", i, c->window, c->name);
+    }
+}
+#endif
 
 static ClientPair
-clientGetTopMostFocusable (ScreenInfo *screen_info, int layer, Client * exclude)
+clientGetTopMostFocusable (ScreenInfo *screen_info, int layer, GList * exclude_list)
 {
     ClientPair top_client;
     Client *c;
@@ -77,9 +98,13 @@ clientGetTopMostFocusable (ScreenInfo *screen_info, int layer, Client * exclude)
             continue;
         }
 
-        if (!exclude || (c != exclude))
+        if (!g_list_find (exclude_list, (gconstpointer) c))
         {
-            if ((c->win_layer <= layer) && FLAG_TEST (c->xfwm_flags, XFWM_FLAG_VISIBLE))
+            if (c->win_layer > layer)
+            {
+                break;
+            }
+            else if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_VISIBLE))
             {
                 if (clientSelectMask (c, 0, WINDOW_REGULAR_FOCUSABLE))
                 {
@@ -87,10 +112,6 @@ clientGetTopMostFocusable (ScreenInfo *screen_info, int layer, Client * exclude)
                 }
                 top_client.highest = c;
             }
-            else if (c->win_layer > layer)
-            {
-                break;
-            }
         }
     }
 
@@ -98,7 +119,7 @@ clientGetTopMostFocusable (ScreenInfo *screen_info, int layer, Client * exclude)
 }
 
 void
-clientFocusTop (ScreenInfo *screen_info, int layer)
+clientFocusTop (ScreenInfo *screen_info, int layer, guint32 timestamp)
 {
     ClientPair top_client;
     DisplayInfo *display_info;
@@ -108,13 +129,13 @@ clientFocusTop (ScreenInfo *screen_info, int layer)
     if (top_client.prefered)
     {
         clientSetFocus (screen_info, top_client.prefered,
-                        myDisplayGetCurrentTime (screen_info->display_info),
+                        timestamp,
                         NO_FOCUS_FLAG);
     }
     else
     {
         clientSetFocus (screen_info, top_client.highest,
-                        myDisplayGetCurrentTime (screen_info->display_info),
+                        timestamp,
                         NO_FOCUS_FLAG);
     }
 }
@@ -143,17 +164,32 @@ clientFocusNew(Client * c)
     }
     else if ((client_focus) && (prevent_focus_stealing))
     {
-        if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_STARTUP_TIME) && (c->user_time == (Time) 0))
+        if (client_focus->win_layer > c->win_layer)
         {
-            TRACE ("Given startup time is 0, not focusing");
+            TRACE ("Not focusing \"%s\" because the current focused window is on a upper layer", c->name);
+            give_focus = FALSE;
+            prevented = TRUE;
+        }
+        else if (client_focus->win_layer < c->win_layer)
+        {
+            /* We short don't use focus stealing prevention against upper layers */
+            TRACE ("Ignoring startup prevention because the current focused window is on a lower layer");
+            give_focus = TRUE;
+            prevented = FALSE;
+        }
+        else if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_STARTUP_TIME) && (c->user_time == CurrentTime))
+        {
+            TRACE ("Given startup time is nil, not focusing \"%s\"", c->name);
             give_focus = FALSE;
             prevented = TRUE;
         }
         else if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_STARTUP_TIME | CLIENT_FLAG_HAS_USER_TIME))
         {
-            if (c->user_time < client_focus->user_time)
+            TRACE ("Current time is %u, time for \"%s\" is %u",
+                   (unsigned int) client_focus->user_time,
+                   c->name, (unsigned int) c->user_time);
+            if (TIMESTAMP_IS_BEFORE (c->user_time, client_focus->user_time))
             {
-                TRACE ("Current %u, new %u", (unsigned int) client_focus->user_time, (unsigned int) c->user_time);
                 give_focus = FALSE;
                 prevented = TRUE;
             }
@@ -172,14 +208,14 @@ clientFocusNew(Client * c)
         clientSetFocus (screen_info, c,
                         myDisplayGetCurrentTime (display_info),
                         FOCUS_IGNORE_MODAL);
+        clientPassGrabMouseButton (c);
     }
     else
     {
         Client *c2 = clientGetFocus();
 
-        if (prevented && c2 && (c2->win_layer == c->win_layer))
+        if ((c2 != NULL) && (c2->win_layer == c->win_layer))
         {
-            FLAG_SET (c->flags, CLIENT_FLAG_DEMANDS_ATTENTION);
             clientSortRing(c);
             clientLower (c, c2->frame);
             clientSortRing(c2);
@@ -191,9 +227,15 @@ clientFocusNew(Client * c)
             clientSortRing(c);
         }
 
+        if (prevented)
+        {
+            TRACE ("clientFocusNew: Setting WM_STATE_DEMANDS_ATTENTION flag on \"%s\" (0x%lx)", c->name, c->window);
+            FLAG_SET (c->flags, CLIENT_FLAG_DEMANDS_ATTENTION);
+        }
+
         clientShow (c, TRUE);
-        clientSetNetState (c);
         clientGrabMouseButton (c);
+        clientSetNetState (c);
     }
 
     return (give_focus);
@@ -282,7 +324,7 @@ clientGetPrevious (Client * c, int mask)
 }
 
 void
-clientPassFocus (ScreenInfo *screen_info, Client *c, Client *exclude)
+clientPassFocus (ScreenInfo *screen_info, Client *c, GList *exclude_list)
 {
     DisplayInfo *display_info;
     ClientPair top_most;
@@ -317,49 +359,20 @@ clientPassFocus (ScreenInfo *screen_info, Client *c, Client *exclude)
     }
 
     display_info = screen_info->display_info;
-    top_most = clientGetTopMostFocusable (screen_info, look_in_layer, exclude);
-    if (screen_info->params->click_to_focus)
+    top_most = clientGetTopMostFocusable (screen_info, look_in_layer, exclude_list);
+
+    if (!(screen_info->params->click_to_focus) &&
+        XQueryPointer (myScreenGetXDisplay (screen_info), screen_info->xroot, &dr, &window, &rx, &ry, &wx, &wy, &mask))
     {
-        if (c)
-        {
-            if (clientIsModal (c))
-            {
-                /* If the window is a modal, send focus back to its parent window.
-                   Modals are transients, and we aren't interested in modal
-                   for group, so it safe to use clientGetTransient because
-                   it's really what we want...
-                 */
-                c2 = clientGetTransient (c);
-                if (c2 && FLAG_TEST(c2->xfwm_flags, XFWM_FLAG_VISIBLE))
-                {
-                    new_focus = c2;
-                    /* Usability: raise the parent, to grab user's attention */
-                    clientRaise (c2, None);
-                }
-            }
-#if 0
-            else
-            {
-                c2 = clientGetNext (c, 0);
-                if ((c2) && (c2->win_layer >= c->win_layer))
-                {
-                    new_focus = c2;
-                }
-            }
-#endif
-        }
-    }
-    else if (XQueryPointer (myScreenGetXDisplay (screen_info), screen_info->xroot, &dr, &window, &rx, &ry, &wx, &wy, &mask))
-    {
-        new_focus = clientAtPosition (screen_info, rx, ry, exclude);
+        new_focus = clientAtPosition (screen_info, rx, ry, exclude_list);
     }
     if (!new_focus)
     {
         new_focus = top_most.prefered ? top_most.prefered : top_most.highest;
     }
     clientSetFocus (screen_info, new_focus,
-                    myDisplayGetCurrentTime (screen_info->display_info),
-                    FOCUS_IGNORE_MODAL | FOCUS_FORCE);
+                    myDisplayGetCurrentTime (display_info),
+                    FOCUS_IGNORE_MODAL | FOCUS_FORCE | FOCUS_TRANSITION);
 }
 
 gboolean
@@ -393,7 +406,7 @@ clientSortRing(Client *c)
     ScreenInfo *screen_info;
 
     g_return_if_fail (c != NULL);
-    TRACE ("Sorting...");
+    TRACE ("entering clientSortRing");
 
     screen_info = c->screen_info;
     if ((screen_info->client_count > 2) && (c != screen_info->clients))
@@ -403,18 +416,66 @@ clientSortRing(Client *c)
 
         c->prev = screen_info->clients->prev;
         c->next = screen_info->clients;
+
         screen_info->clients->prev->next = c;
         screen_info->clients->prev = c;
     }
     screen_info->clients = c;
 }
 
+void
+clientSetLast(Client *c)
+{
+    ScreenInfo *screen_info;
+
+    g_return_if_fail (c != NULL);
+    TRACE ("entering clientSetLast");
+
+    screen_info = c->screen_info;
+    if (screen_info->client_count > 2)
+    {
+        if (screen_info->clients == c)
+        {
+            screen_info->clients = screen_info->clients->next;
+        }
+        else
+        {
+            c->prev->next = c->next;
+            c->next->prev = c->prev;
+
+            c->prev = screen_info->clients->prev;
+            c->next = screen_info->clients;
+
+            screen_info->clients->prev->next = c;
+            screen_info->clients->prev = c;
+        }
+    }
+}
+
+static void
+clientFocusNone (ScreenInfo *screen_info, Client *previous, guint32 timestamp)
+{
+    TRACE ("setting focus to none");
+
+    pending_focus = NULL;
+
+    if (previous)
+    {
+        frameQueueDraw (previous, FALSE);
+        if (previous->screen_info != screen_info)
+        {
+            clientSetNetActiveWindow (previous->screen_info, NULL, timestamp);
+        }
+    }
+    clientSetNetActiveWindow (screen_info, NULL, timestamp);
+    XSetInputFocus (myScreenGetXDisplay (screen_info), screen_info->xfwm4_win, RevertToPointerRoot, timestamp);
+}
+
 void
 clientUpdateFocus (ScreenInfo *screen_info, Client * c, unsigned short flags)
 {
     DisplayInfo *display_info;
     Client *c2;
-    unsigned long data[2];
 
     TRACE ("entering clientUpdateFocus");
 
@@ -445,38 +506,37 @@ clientUpdateFocus (ScreenInfo *screen_info, Client * c, unsigned short flags)
     client_focus = c;
     if (c)
     {
+        user_focus = c;
         clientInstallColormaps (c);
         if (flags & FOCUS_SORT)
         {
             clientSortRing(c);
         }
-        data[0] = c->window;
+        if (FLAG_TEST(c->flags, CLIENT_FLAG_DEMANDS_ATTENTION))
+        {
+            TRACE ("Un-setting WM_STATE_DEMANDS_ATTENTION flag on \"%s\" (0x%lx)", c->name, c->window);
+            FLAG_UNSET (c->flags, CLIENT_FLAG_DEMANDS_ATTENTION);
+        }
         clientAdjustFullscreenLayer (c, TRUE);
-        frameDraw (c, FALSE);
-    }
-    else
-    {
-        data[0] = None;
+        frameQueueDraw (c, FALSE);
+        clientUpdateOpacity (c);
+        clientSetNetState (c);
     }
     if (c2)
     {
         if (c)
         {
             clientAdjustFullscreenLayer (c2, FALSE);
-            /* clientRaise (c, None); */
         }
-        frameDraw (c2, FALSE);
+        frameQueueDraw (c2, FALSE);
+        clientUpdateOpacity (c2);
     }
-    data[1] = None;
-    XChangeProperty (display_info->dpy, screen_info->xroot,
-                     display_info->atoms[NET_ACTIVE_WINDOW], XA_WINDOW, 32,
-                     PropModeReplace, (unsigned char *) data, 2);
-
-    clientUpdateOpacity (screen_info, c);
+    clientSetNetActiveWindow (screen_info, c, 0);
+    clientClearDelayedFocus ();
 }
 
 void
-clientSetFocus (ScreenInfo *screen_info, Client * c, Time timestamp, unsigned short flags)
+clientSetFocus (ScreenInfo *screen_info, Client *c, guint32 timestamp, unsigned short flags)
 {
     DisplayInfo *display_info;
     Client *c2;
@@ -497,16 +557,17 @@ clientSetFocus (ScreenInfo *screen_info, Client * c, Time timestamp, unsigned sh
     c2 = ((client_focus != c) ? client_focus : NULL);
     if ((c) && FLAG_TEST (c->xfwm_flags, XFWM_FLAG_VISIBLE))
     {
-        TRACE ("setting focus to client \"%s\" (0x%lx)", c->name, c->window);
+        TRACE ("setting focus to client \"%s\" (0x%lx) with timestamp %u", c->name, c->window, (unsigned int) timestamp);
+        user_focus = c;
         if (FLAG_TEST(c->flags, CLIENT_FLAG_DEMANDS_ATTENTION))
         {
+            TRACE ("Un-setting WM_STATE_DEMANDS_ATTENTION flag on \"%s\" (0x%lx)", c->name, c->window);
             FLAG_UNSET (c->flags, CLIENT_FLAG_DEMANDS_ATTENTION);
             clientSetNetState (c);
         }
         if ((c == client_focus) && !(flags & FOCUS_FORCE))
         {
-            TRACE ("client \"%s\" (0x%lx) is already focused, ignoring request",
-                c->name, c->window);
+            TRACE ("client \"%s\" (0x%lx) is already focused, ignoring request", c->name, c->window);
             return;
         }
         if (!clientAcceptFocus (c))
@@ -529,31 +590,32 @@ clientSetFocus (ScreenInfo *screen_info, Client * c, Time timestamp, unsigned sh
             {
                 XSetInputFocus (myScreenGetXDisplay (screen_info), c->window, RevertToPointerRoot, timestamp);
             }
-            clientUpdateOpacity (screen_info, c);
         }
+        else if (flags & FOCUS_TRANSITION)
+        {
+            /*
+             * If we are relying only on the client application to take focus, we need to set the focus
+             * explicitely on our own fallback window otherwise there is a race condition between the
+             * application and the window manager. If the application does not take focus before the
+             * the previously focused window is unmapped (when iconifying or closing for example), the focus
+             * will be reverted to the root window and focus transition will fail.
+             */
+            clientFocusNone (screen_info, c2, timestamp);
+        }
+
         if (FLAG_TEST(c->wm_flags, WM_FLAG_TAKEFOCUS))
         {
+            pending_focus = c;
             sendClientMessage (c->screen_info, c->window, WM_TAKE_FOCUS, timestamp);
         }
     }
     else
     {
-        unsigned long data[2];
-
         TRACE ("setting focus to none");
 
-        data[0] = data[1] = None;
         client_focus = NULL;
-        if (c2)
-        {
-            frameDraw (c2, FALSE);
-            XChangeProperty (clientGetXDisplay (c2), c2->screen_info->xroot, display_info->atoms[NET_ACTIVE_WINDOW], XA_WINDOW, 32,
-                             PropModeReplace, (unsigned char *) data, 2);
-        }
-        XChangeProperty (myScreenGetXDisplay (screen_info), screen_info->xroot, display_info->atoms[NET_ACTIVE_WINDOW], XA_WINDOW, 32,
-                         PropModeReplace, (unsigned char *) data, 2);
-        XSetInputFocus (myScreenGetXDisplay (screen_info), screen_info->xfwm4_win, RevertToPointerRoot, timestamp);
-        clientUpdateOpacity (screen_info, c);
+        clientFocusNone (screen_info, c2, timestamp);
+        clientClearDelayedFocus ();
     }
 }
 
@@ -592,10 +654,43 @@ clientGetFocus (void)
     return (client_focus);
 }
 
-void
-clientClearFocus (void)
+Client *
+clientGetFocusPending (void)
 {
-    client_focus = NULL;
+    return (pending_focus);
+}
+
+Client *
+clientGetFocusOrPending (void)
+{
+    if (client_focus)
+    {
+        return (client_focus);
+    }
+    return (pending_focus);
+}
+
+Client *
+clientGetUserFocus (void)
+{
+    return (user_focus);
+}
+
+void
+clientClearFocus (Client *c)
+{
+    if ((c == NULL) || (c == client_focus))
+    {
+        client_focus = NULL;
+    }
+    if ((c == NULL) || (c == pending_focus))
+    {
+        pending_focus = NULL;
+    }
+    if ((c == NULL) || (c == user_focus))
+    {
+        user_focus = NULL;
+    }
 }
 
 void
@@ -680,17 +775,16 @@ clientPassGrabMouseButton (Client * c)
         return;
     }
 
-    TRACE ("ungrabing buttons for client \"%s\" (0x%lx)", c->name, c->window);
     if (last_ungrab == c)
     {
         return;
     }
 
+    TRACE ("ungrabing buttons for client \"%s\" (0x%lx)", c->name, c->window);
     if (last_ungrab)
     {
         clientGrabMouseButton (last_ungrab);
     }
-
     clientUngrabMouseButton (c);
     last_ungrab = c;
 }
@@ -706,3 +800,51 @@ clientClearLastUngrab (void)
 {
     last_ungrab = NULL;
 }
+
+static gboolean
+delayed_focus_cb (gpointer data)
+{
+    ScreenInfo *screen_info;
+    DisplayInfo *display_info;
+
+    TRACE ("entering delayed_focus_cb");
+    g_return_val_if_fail (delayed_focus != NULL, FALSE);
+
+    screen_info = delayed_focus->screen_info;
+    display_info = screen_info->display_info;
+    clientSetFocus (screen_info, delayed_focus, myDisplayGetCurrentTime (display_info), NO_FOCUS_FLAG);
+    focus_timeout = 0;
+    delayed_focus = NULL;
+
+    return (FALSE);
+}
+
+void
+clientClearDelayedFocus (void)
+{
+    if(focus_timeout)
+    {
+        g_source_remove (focus_timeout);
+        focus_timeout = 0;
+    }
+    delayed_focus = NULL;
+}
+
+void
+clientAddDelayedFocus (Client *c)
+{
+    ScreenInfo *screen_info;
+
+    screen_info = c->screen_info;
+    delayed_focus = c;
+    focus_timeout = g_timeout_add_full (G_PRIORITY_DEFAULT,
+                                        screen_info->params->focus_delay,
+                                        (GSourceFunc) delayed_focus_cb,
+                                        NULL, NULL);
+}
+
+Client *
+clientGetDelayedFocus (void)
+{
+    return delayed_focus;
+}
