@@ -1,4 +1,4 @@
-/*      $Id: netwm.c 24346 2007-01-10 20:50:18Z olivier $
+/*      $Id: netwm.c 29542 2009-02-23 10:06:52Z olivier $
 
         This program is free software; you can redistribute it and/or modify
         it under the terms of the GNU General Public License as published by
@@ -12,14 +12,16 @@
 
         You should have received a copy of the GNU General Public License
         along with this program; if not, write to the Free Software
-        Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+        Foundation, Inc., Inc., 51 Franklin Street, Fifth Floor, Boston,
+        MA 02110-1301, USA.
 
-        xfwm4    - (c) 2002-2006 Olivier Fourdan
+
+        xfwm4    - (c) 2002-2009 Olivier Fourdan
 
  */
 
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+#include "config.h"
 #endif
 
 #include <X11/Xlib.h>
@@ -32,6 +34,7 @@
 #include <libxfce4util/libxfce4util.h>
 
 #include "client.h"
+#include "moveresize.h"
 #include "compositor.h"
 #include "display.h"
 #include "frame.h"
@@ -41,6 +44,7 @@
 #include "netwm.h"
 #include "screen.h"
 #include "stacking.h"
+#include "terminate.h"
 #include "transients.h"
 #include "workspaces.h"
 
@@ -85,7 +89,7 @@ clientSetNetState (Client * c)
         TRACE ("clientSetNetState : skip_taskbar");
         data[i++] = display_info->atoms[NET_WM_STATE_SKIP_TASKBAR];
     }
-    if (FLAG_TEST (c->flags, CLIENT_FLAG_MAXIMIZED))
+    if (FLAG_TEST_ALL (c->flags, CLIENT_FLAG_MAXIMIZED))
     {
         TRACE ("clientSetNetState : maximize vert + horiz");
         data[i++] = display_info->atoms[NET_WM_STATE_MAXIMIZED_HORZ];
@@ -264,7 +268,10 @@ clientGetNetState (Client * c)
             }
             else
             {
-                g_warning ("Unmanaged net_wm_state (window 0x%lx)", c->window);
+                gchar *atom_name;
+                atom_name = get_atom_name (display_info, atoms[i]);
+                g_warning ("Unmanaged net_wm_state (window 0x%lx, atom \"%s\")", c->window, atom_name);
+                g_free (atom_name);
             }
 
             ++i;
@@ -317,7 +324,7 @@ clientUpdateNetState (Client * c, XClientMessageEvent * ev)
     if ((first  == display_info->atoms[NET_WM_STATE_STICKY]) ||
         (second == display_info->atoms[NET_WM_STATE_STICKY]))
     {
-        if (!clientIsValidTransientOrModal (c) && FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_STICK))
+        if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_STICK))
         {
             if ((action == NET_WM_STATE_ADD) && !FLAG_TEST (c->flags, CLIENT_FLAG_STICKY))
             {
@@ -331,7 +338,7 @@ clientUpdateNetState (Client * c, XClientMessageEvent * ev)
             {
                 clientToggleSticky (c, TRUE);
             }
-            frameDraw (c, FALSE);
+            frameQueueDraw (c, FALSE);
         }
     }
 
@@ -396,22 +403,22 @@ clientUpdateNetState (Client * c, XClientMessageEvent * ev)
         if ((action == NET_WM_STATE_ADD) && !FLAG_TEST (c->flags, CLIENT_FLAG_STATE_MODAL))
         {
             FLAG_SET (c->flags, CLIENT_FLAG_STATE_MODAL);
-            clientSetNetState (c);
             clientWindowType (c);
+            clientSetNetState (c);
         }
         else if ((action == NET_WM_STATE_REMOVE) && FLAG_TEST (c->flags, CLIENT_FLAG_STATE_MODAL))
         {
             FLAG_UNSET (c->flags, CLIENT_FLAG_STATE_MODAL);
-            clientSetNetState (c);
             clientWindowType (c);
+            clientSetNetState (c);
         }
         else if (action == NET_WM_STATE_TOGGLE)
         {
             FLAG_TOGGLE (c->flags, CLIENT_FLAG_STATE_MODAL);
-            clientSetNetState (c);
             clientWindowType (c);
+            clientSetNetState (c);
         }
-        frameDraw (c, TRUE);
+        frameQueueDraw (c, TRUE);
     }
 
     if ((first  == display_info->atoms[NET_WM_STATE_FULLSCREEN]) ||
@@ -445,17 +452,17 @@ clientUpdateNetState (Client * c, XClientMessageEvent * ev)
             if ((action == NET_WM_STATE_ADD) && !FLAG_TEST (c->flags, CLIENT_FLAG_ABOVE))
             {
                 FLAG_SET (c->flags, CLIENT_FLAG_ABOVE);
-                clientUpdateAboveState (c);
+                clientUpdateLayerState (c);
             }
             else if ((action == NET_WM_STATE_REMOVE) && FLAG_TEST (c->flags, CLIENT_FLAG_ABOVE))
             {
                 FLAG_UNSET (c->flags, CLIENT_FLAG_ABOVE);
-                clientUpdateAboveState (c);
+                clientUpdateLayerState (c);
             }
             else if (action == NET_WM_STATE_TOGGLE)
             {
                 FLAG_TOGGLE (c->flags, CLIENT_FLAG_ABOVE);
-                clientUpdateAboveState (c);
+                clientUpdateLayerState (c);
             }
         }
     }
@@ -467,21 +474,18 @@ clientUpdateNetState (Client * c, XClientMessageEvent * ev)
         {
             if ((action == NET_WM_STATE_ADD) && !FLAG_TEST (c->flags, CLIENT_FLAG_BELOW))
             {
-                compositorDamageWindow (display_info, c->frame);
                 FLAG_SET (c->flags, CLIENT_FLAG_BELOW);
-                clientUpdateBelowState (c);
+                clientUpdateLayerState (c);
             }
             else if ((action == NET_WM_STATE_REMOVE) && FLAG_TEST (c->flags, CLIENT_FLAG_BELOW))
             {
-                compositorDamageWindow (display_info, c->frame);
                 FLAG_UNSET (c->flags, CLIENT_FLAG_BELOW);
-                clientUpdateBelowState (c);
+                clientUpdateLayerState (c);
             }
             else if (action == NET_WM_STATE_TOGGLE)
             {
-                compositorDamageWindow (display_info, c->frame);
                 FLAG_TOGGLE (c->flags, CLIENT_FLAG_BELOW);
-                clientUpdateBelowState (c);
+                clientUpdateLayerState (c);
             }
         }
     }
@@ -524,7 +528,7 @@ clientUpdateNetState (Client * c, XClientMessageEvent * ev)
             FLAG_TOGGLE (c->flags, CLIENT_FLAG_SKIP_TASKBAR);
             clientSetNetState (c);
         }
-        frameDraw (c, TRUE);
+        frameQueueDraw (c, TRUE);
     }
 
     if ((first  == display_info->atoms[NET_WM_STATE_DEMANDS_ATTENTION]) ||
@@ -597,8 +601,8 @@ clientNetMoveResize (Client * c, XClientMessageEvent * ev)
         }
         else
         {
-            g_warning ("Could not determine the mouse button used");
-            return;
+            /* Fallback */
+            button = Button1;
         }
     }
 
@@ -608,7 +612,7 @@ clientNetMoveResize (Client * c, XClientMessageEvent * ev)
     event->xbutton.button = button;
     event->xbutton.x_root = event->xkey.x_root = x_root;
     event->xbutton.y_root = event->xkey.y_root = y_root;
-    event->xbutton.time = event->xkey.time = myDisplayGetCurrentTime (display_info);
+    event->xbutton.time = event->xkey.time = (Time) myDisplayGetCurrentTime (display_info);
 
     switch (action)
     {
@@ -672,7 +676,7 @@ clientNetMoveResize (Client * c, XClientMessageEvent * ev)
             resize = FALSE; /* Move */
             break;
     }
-    
+
     if (!FLAG_TEST (c->flags, CLIENT_FLAG_FULLSCREEN))
     {
         if (resize && FLAG_TEST_ALL (c->xfwm_flags, XFWM_FLAG_HAS_RESIZE | XFWM_FLAG_IS_RESIZABLE))
@@ -686,12 +690,44 @@ clientNetMoveResize (Client * c, XClientMessageEvent * ev)
     }
 }
 
+void
+clientNetMoveResizeWindow (Client * c, XClientMessageEvent * ev)
+{
+    XWindowChanges wc;
+    unsigned long mask;
+    int gravity;
+
+    g_return_if_fail (c != NULL);
+    TRACE ("entering clientNetMoveResizeWindow");
+    TRACE ("client \"%s\" (0x%lx)", c->name, c->window);
+
+    if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_MOVING_RESIZING))
+    {
+        /* not allowed */
+        return;
+    }
+
+    gravity = (ev->data.l[0] & 0xff);
+    if (!gravity)
+    {
+        gravity = c->gravity;
+    }
+
+    wc.x = ev->data.l[1];
+    wc.y = ev->data.l[2];
+    wc.width = ev->data.l[3];
+    wc.height = ev->data.l[4];
+    mask = (ev->data.l[0] & 0xf00) >> 8;
+
+    clientAdjustCoordGravity (c, gravity, &mask, &wc);
+    clientMoveResizeWindow (c, &wc, mask);
+}
+
 void
 clientUpdateFullscreenState (Client * c)
 {
     ScreenInfo *screen_info;
     DisplayInfo *display_info;
-    XWindowChanges wc;
     int layer;
 
     g_return_if_fail (c != NULL);
@@ -703,57 +739,28 @@ clientUpdateFullscreenState (Client * c)
 
     if (FLAG_TEST (c->flags, CLIENT_FLAG_FULLSCREEN))
     {
-        GdkRectangle rect;
-        gint monitor_nbr;
-        int cx, cy;
-
-        cx = frameX (c) + (frameWidth (c) / 2);
-        cy = frameY (c) + (frameHeight (c) / 2);
-
-        monitor_nbr = find_monitor_at_point (c->screen_info->gscr, cx, cy);
-        gdk_screen_get_monitor_geometry (c->screen_info->gscr, monitor_nbr, &rect);
-
         c->fullscreen_old_x = c->x;
         c->fullscreen_old_y = c->y;
         c->fullscreen_old_width = c->width;
         c->fullscreen_old_height = c->height;
         c->fullscreen_old_layer = c->win_layer;
-
-        wc.x = rect.x;
-        wc.y = rect.y;
-        wc.width = rect.width;
-        wc.height = rect.height;
-        layer = WIN_LAYER_ABOVE_DOCK;
+        layer = WIN_LAYER_FULLSCREEN;
     }
     else
     {
-        wc.x = c->fullscreen_old_x;
-        wc.y = c->fullscreen_old_y;
-        wc.width = c->fullscreen_old_width;
-        wc.height = c->fullscreen_old_height;
         layer = c->fullscreen_old_layer;
     }
-    clientSetNetState (c);
-    if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_MANAGED))
-    {
-        /*
-           For some reason, the configure can generate EnterNotify events
-           on lower windows, causing a nasty race cond with apps trying to
-           grab focus in focus follow mouse mode. Grab the pointer to
-           avoid these effects
-         */
-        myScreenGrabPointer (c->screen_info, EnterWindowMask, None, myDisplayGetCurrentTime (display_info));
-        clientConfigure (c, &wc, CWX | CWY | CWWidth | CWHeight, CFG_FORCE_REDRAW);
-        myScreenUngrabPointer (c->screen_info);
-    }
-    else
-    {
-        c->x = wc.x;
-        c->y = wc.y;
-        c->height = wc.height;
-        c->width = wc.width;
-    }
     clientSetLayer (c, layer);
+    clientUpdateFullscreenSize (c);
+
+    /* Fullscreen has no decoration at all, update NET_FRAME_EXTENTS accordingly */
+    setNetFrameExtents (display_info,
+                        c->window,
+                        frameTop (c),
+                        frameLeft (c),
+                        frameRight (c),
+                        frameBottom (c));
+    clientSetNetState (c);
 }
 
 void
@@ -773,6 +780,7 @@ clientGetNetWmType (Client * c)
     display_info = screen_info->display_info;
     n_atoms = 0;
     atoms = NULL;
+    c->type_atom = None;
 
     if (!getAtomList (display_info, c->window, NET_WM_WINDOW_TYPE, &atoms, &n_atoms))
     {
@@ -788,7 +796,14 @@ clientGetNetWmType (Client * c)
                 c->type_atom = display_info->atoms[NET_WM_WINDOW_TYPE_NORMAL];
                 break;
             default:
-                c->type_atom = None;
+                if (c->transient_for != None)
+                {
+                    c->type_atom = display_info->atoms[NET_WM_WINDOW_TYPE_DIALOG];
+                }
+                else
+                {
+                    c->type_atom = display_info->atoms[NET_WM_WINDOW_TYPE_NORMAL];
+                }
                 break;
         }
     }
@@ -835,53 +850,56 @@ clientGetInitialNetWmDesktop (Client * c)
     display_info = screen_info->display_info;
     val = 0;
 
-    /* This is to make sure that transient are shown with their "ancestor" window */
-    c2 = clientGetTransient (c);
-    if (c2)
+    if (!FLAG_TEST (c->xfwm_flags, XFWM_FLAG_SESSION_MANAGED)
+        &&  !FLAG_TEST (c->xfwm_flags, XFWM_FLAG_WORKSPACE_SET))
     {
         FLAG_SET (c->xfwm_flags, XFWM_FLAG_WORKSPACE_SET);
-        c->win_workspace = c2->win_workspace;
-        if (FLAG_TEST (c2->flags, CLIENT_FLAG_STICKY))
-        {
-            FLAG_SET (c->flags, CLIENT_FLAG_STICKY);
-            c->win_state |= WIN_STATE_STICKY;
-        }
+        c->win_workspace = c->screen_info->current_ws;
     }
-    else
+    if (getHint (display_info, c->window, NET_WM_DESKTOP, &val))
     {
-        if (!FLAG_TEST (c->xfwm_flags, XFWM_FLAG_SESSION_MANAGED)
-            &&  !FLAG_TEST (c->xfwm_flags, XFWM_FLAG_WORKSPACE_SET))
+        TRACE ("atom net_wm_desktop detected");
+        if (val == (int) ALL_WORKSPACES)
         {
-            FLAG_SET (c->xfwm_flags, XFWM_FLAG_WORKSPACE_SET);
+            if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_STICK) && !FLAG_TEST (c->flags, CLIENT_FLAG_STICKY))
+            {
+                TRACE ("atom net_wm_desktop specifies window \"%s\" is sticky", c->name);
+                FLAG_SET (c->flags, CLIENT_FLAG_STICKY);
+                c->win_state |= WIN_STATE_STICKY;
+            }
             c->win_workspace = c->screen_info->current_ws;
         }
-        if (getHint (display_info, c->window, NET_WM_DESKTOP, &val))
+        else
         {
-            TRACE ("atom net_wm_desktop detected");
-            if (val == (int) ALL_WORKSPACES)
+            TRACE ("atom net_wm_desktop specifies window \"%s\" is on desk %i", c->name, (int) val);
+            c->win_workspace = (int) val;
+        }
+        FLAG_SET (c->xfwm_flags, XFWM_FLAG_WORKSPACE_SET);
+    }
+    else if (getHint (display_info, c->window, WIN_WORKSPACE, &val))
+    {
+        TRACE ("atom win_workspace specifies window \"%s\" is on desk %i", c->name, (int) val);
+        c->win_workspace = (int) val;
+        FLAG_SET (c->xfwm_flags, XFWM_FLAG_WORKSPACE_SET);
+    }
+
+    /* This is to make sure that transient are shown with their "ancestor" window */
+    if (!FLAG_TEST (c->flags, CLIENT_FLAG_STICKY))
+    {
+        c2 = clientGetTransient (c);
+        if (c2)
+        {
+            FLAG_SET (c->xfwm_flags, XFWM_FLAG_WORKSPACE_SET);
+            c->win_workspace = c2->win_workspace;
+            if (FLAG_TEST (c2->flags, CLIENT_FLAG_STICKY))
             {
-                if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_STICK) && !FLAG_TEST (c->flags, CLIENT_FLAG_STICKY))
-                {
-                    TRACE ("atom net_wm_desktop specifies window \"%s\" is sticky", c->name);
-                    FLAG_SET (c->flags, CLIENT_FLAG_STICKY);
-                    c->win_state |= WIN_STATE_STICKY;
-                }
+                FLAG_SET (c->flags, CLIENT_FLAG_STICKY);
+                c->win_state |= WIN_STATE_STICKY;
                 c->win_workspace = c->screen_info->current_ws;
             }
-            else
-            {
-                TRACE ("atom net_wm_desktop specifies window \"%s\" is on desk %i", c->name, (int) val);
-                c->win_workspace = (int) val;
-            }
-            FLAG_SET (c->xfwm_flags, XFWM_FLAG_WORKSPACE_SET);
-        }
-        else if (getHint (display_info, c->window, WIN_WORKSPACE, &val))
-        {
-            TRACE ("atom win_workspace specifies window \"%s\" is on desk %i", c->name, (int) val);
-            c->win_workspace = (int) val;
-            FLAG_SET (c->xfwm_flags, XFWM_FLAG_WORKSPACE_SET);
         }
     }
+
     TRACE ("initial desktop for window \"%s\" is %i", c->name, c->win_workspace);
     if (c->win_workspace > c->screen_info->workspace_count - 1)
     {
@@ -940,7 +958,6 @@ clientValidateNetStrut (Client * c)
     ScreenInfo *screen_info;
     gboolean valid;
     int max_value;
-    int i;
 
     g_return_val_if_fail (c != NULL, TRUE);
     TRACE ("entering clientValidateNetStrut for \"%s\" (0x%lx)", c->name, c->window);
@@ -948,14 +965,31 @@ clientValidateNetStrut (Client * c)
     max_value = MIN (screen_info->width, screen_info->height) / 4;
     valid = TRUE;
 
-    for (i = 0; i < 4; i++)
+    if (c->struts[STRUTS_TOP] > screen_info->height - screen_info->margins[STRUTS_BOTTOM])
     {
-        if (c->struts[i] > max_value)
-        {
-           g_warning ("Strut value for application window 0x%lx changed from %d to %d", c->window, c->struts[i], max_value);
-           c->struts[i] = max_value;
-           valid = FALSE;
-        }
+       c->struts[STRUTS_TOP] = screen_info->height - screen_info->margins[STRUTS_BOTTOM];
+       g_warning ("Top strut value for application window 0x%lx confined to %d", c->window, c->struts[STRUTS_TOP]);
+       valid = FALSE;
+    }
+
+    if (c->struts[STRUTS_BOTTOM] > screen_info->height - screen_info->margins[STRUTS_TOP])
+    {
+       c->struts[STRUTS_BOTTOM] = screen_info->height - screen_info->margins[STRUTS_TOP];
+       g_warning ("Bottom strut value for application window 0x%lx confined to %d", c->window, c->struts[STRUTS_BOTTOM]);
+       valid = FALSE;
+    }
+
+    if (c->struts[STRUTS_LEFT] > screen_info->width - screen_info->margins[STRUTS_RIGHT])
+    {
+       c->struts[STRUTS_LEFT] = screen_info->height - screen_info->margins[STRUTS_RIGHT];
+       g_warning ("Left strut value for application window 0x%lx confined to %d", c->window, c->struts[STRUTS_LEFT]);
+       valid = FALSE;
+    }
+    if (c->struts[STRUTS_RIGHT] > screen_info->width - screen_info->margins[STRUTS_LEFT])
+    {
+       c->struts[STRUTS_RIGHT] = screen_info->height - screen_info->margins[STRUTS_LEFT];
+       g_warning ("Right strut value for application window 0x%lx confined to %d", c->window, c->struts[STRUTS_RIGHT]);
+       valid = FALSE;
     }
 
     return valid;
@@ -1090,7 +1124,7 @@ clientSetNetActions (Client * c)
 {
     ScreenInfo *screen_info;
     DisplayInfo *display_info;
-    Atom atoms[6];
+    Atom atoms[16];
     int i;
 
     g_return_if_fail (c != NULL);
@@ -1100,21 +1134,43 @@ clientSetNetActions (Client * c)
     display_info = screen_info->display_info;
     i = 0;
 
+    /* Actions available for all */
     atoms[i++] = display_info->atoms[NET_WM_ACTION_CLOSE];
-    if (CLIENT_CAN_MAXIMIZE_WINDOW (c))
+    atoms[i++] = display_info->atoms[NET_WM_ACTION_ABOVE];
+    atoms[i++] = display_info->atoms[NET_WM_ACTION_BELOW];
+
+    /* Actions depending on the window type and current status */
+    if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_VISIBLE))
     {
-        atoms[i++] = display_info->atoms[NET_WM_ACTION_MAXIMIZE_HORZ];
-        atoms[i++] = display_info->atoms[NET_WM_ACTION_MAXIMIZE_VERT];
+        atoms[i++] = display_info->atoms[NET_WM_ACTION_FULLSCREEN];
+        if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_MOVE))
+        {
+            atoms[i++] = display_info->atoms[NET_WM_ACTION_MOVE];
+        }
+            if (FLAG_TEST_ALL (c->xfwm_flags, XFWM_FLAG_HAS_RESIZE | XFWM_FLAG_IS_RESIZABLE) &&
+                !FLAG_TEST_ALL (c->flags, CLIENT_FLAG_MAXIMIZED))
+        {
+            atoms[i++] = display_info->atoms[NET_WM_ACTION_RESIZE];
+        }
+        if (CLIENT_CAN_MAXIMIZE_WINDOW (c))
+        {
+            atoms[i++] = display_info->atoms[NET_WM_ACTION_MAXIMIZE_HORZ];
+            atoms[i++] = display_info->atoms[NET_WM_ACTION_MAXIMIZE_VERT];
+        }
+        if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_BORDER))
+        {
+            atoms[i++] = display_info->atoms[NET_WM_ACTION_SHADE];
+        }
+    }
+    if (CLIENT_CAN_HIDE_WINDOW (c))
+    {
+        atoms[i++] = display_info->atoms[NET_WM_ACTION_MINIMIZE];
     }
     if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_STICK))
     {
         atoms[i++] = display_info->atoms[NET_WM_ACTION_CHANGE_DESKTOP];
         atoms[i++] = display_info->atoms[NET_WM_ACTION_STICK];
     }
-    if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_BORDER))
-    {
-        atoms[i++] = display_info->atoms[NET_WM_ACTION_SHADE];
-    }
     XChangeProperty (clientGetXDisplay (c), c->window, display_info->atoms[NET_WM_ALLOWED_ACTIONS],
                      XA_ATOM, 32, PropModeReplace, (unsigned char *) atoms, i);
 }
@@ -1157,8 +1213,9 @@ clientWindowType (Client * c)
             TRACE ("atom net_wm_window_type_dock detected");
             c->type = WINDOW_DOCK;
             c->initial_layer = WIN_LAYER_DOCK;
-            FLAG_SET (c->flags,
-                CLIENT_FLAG_SKIP_PAGER | CLIENT_FLAG_SKIP_TASKBAR);
+             FLAG_SET (c->flags,
+                CLIENT_FLAG_SKIP_PAGER | CLIENT_FLAG_STICKY |
+                CLIENT_FLAG_SKIP_TASKBAR);
             FLAG_UNSET (c->xfwm_flags,
                 XFWM_FLAG_HAS_BORDER |  XFWM_FLAG_HAS_MOVE |
                 XFWM_FLAG_HAS_HIDE | XFWM_FLAG_HAS_MAXIMIZE |
@@ -1193,6 +1250,11 @@ clientWindowType (Client * c)
             TRACE ("atom net_wm_window_type_dialog detected");
             c->type = WINDOW_DIALOG;
             c->initial_layer = WIN_LAYER_NORMAL;
+            /* Treat DIALOG without transient_for set as transient for group */
+            if (c->transient_for == None)
+            {
+                c->transient_for = c->screen_info->xroot;
+            }
         }
         else if (c->type_atom == display_info->atoms[NET_WM_WINDOW_TYPE_NORMAL])
         {
@@ -1207,12 +1269,17 @@ clientWindowType (Client * c)
                 CLIENT_FLAG_SKIP_PAGER | CLIENT_FLAG_SKIP_TASKBAR);
             c->type = WINDOW_UTILITY;
             c->initial_layer = WIN_LAYER_NORMAL;
+            /* Treat UTILITY without transient_for set as transient for group */
+            if (c->transient_for == None)
+            {
+                c->transient_for = c->screen_info->xroot;
+            }
         }
         else if (c->type_atom == display_info->atoms[NET_WM_WINDOW_TYPE_SPLASH])
         {
             TRACE ("atom net_wm_window_type_splash detected");
             c->type = WINDOW_SPLASHSCREEN;
-            c->initial_layer = WIN_LAYER_ABOVE_DOCK;
+            c->initial_layer = WIN_LAYER_NORMAL;
             FLAG_SET (c->flags,
                 CLIENT_FLAG_SKIP_PAGER | CLIENT_FLAG_SKIP_TASKBAR);
             FLAG_UNSET (c->xfwm_flags,
@@ -1240,48 +1307,30 @@ clientWindowType (Client * c)
             c->initial_layer = c2->win_layer;
             TRACE ("Applied layer is %i", c->initial_layer);
         }
-        FLAG_UNSET (c->xfwm_flags, XFWM_FLAG_HAS_HIDE | XFWM_FLAG_HAS_STICK);
-        FLAG_UNSET (c->flags, CLIENT_FLAG_STICKY);
+        FLAG_UNSET (c->xfwm_flags, XFWM_FLAG_HAS_HIDE);
     }
     if ((old_type != c->type) || (c->initial_layer != c->win_layer))
     {
         TRACE ("setting layer %i", c->initial_layer);
-        clientSetNetState (c);
         clientSetLayer (c, c->initial_layer);
+        clientSetNetState (c);
     }
 }
 
 void
-clientUpdateAboveState (Client * c)
+clientUpdateLayerState (Client * c)
 {
     int layer;
 
     g_return_if_fail (c != NULL);
-    TRACE ("entering clientUpdateAboveState");
-    TRACE ("Update above state for client \"%s\" (0x%lx)", c->name, c->window);
+    TRACE ("entering clientUpdateLayerState");
+    TRACE ("Update layer state for client \"%s\" (0x%lx)", c->name, c->window);
 
     if (FLAG_TEST (c->flags, CLIENT_FLAG_ABOVE))
     {
         layer = WIN_LAYER_ABOVE_DOCK;
     }
-    else
-    {
-        layer = c->initial_layer;
-    }
-    clientSetNetState (c);
-    clientSetLayer (c, layer);
-}
-
-void
-clientUpdateBelowState (Client * c)
-{
-    int layer;
-
-    g_return_if_fail (c != NULL);
-    TRACE ("entering clientUpdateBelowState");
-    TRACE ("Update below state for client \"%s\" (0x%lx)", c->name, c->window);
-
-    if (FLAG_TEST (c->flags, CLIENT_FLAG_BELOW))
+    else if (FLAG_TEST (c->flags, CLIENT_FLAG_BELOW))
     {
         layer = WIN_LAYER_BELOW;
     }
@@ -1289,7 +1338,176 @@ clientUpdateBelowState (Client * c)
     {
         layer = c->initial_layer;
     }
-    clientSetNetState (c);
     clientSetLayer (c, layer);
+    clientSetNetState (c);
 }
 
+void
+clientSetNetActiveWindow (ScreenInfo *screen_info, Client *c, guint32 timestamp)
+{
+    DisplayInfo *display_info;
+    unsigned long data[2];
+
+    g_return_if_fail (screen_info != NULL);
+    TRACE ("entering clientSetNetActiveWindow");
+
+    display_info = screen_info->display_info;
+    data[0] = (unsigned long) None;
+    data[1] = (unsigned long) timestamp;
+    if (c)
+    {
+        data[0] = (unsigned long) c->window;
+    }
+    XChangeProperty (myScreenGetXDisplay (screen_info), screen_info->xroot, display_info->atoms[NET_ACTIVE_WINDOW], XA_WINDOW, 32,
+                     PropModeReplace, (unsigned char *) data, 2);
+}
+
+static gboolean
+ping_timeout_cb (gpointer data)
+{
+    Client *c;
+
+    TRACE ("entering ping_timeout_cb");
+
+    c = (Client *) data;
+    if (c)
+    {
+        c->ping_timeout_id = 0;
+        TRACE ("Ping timeout on client \"%s\"", c->name);
+        terminateShowDialog (c);
+    }
+    return (FALSE);
+}
+
+void
+clientRemoveNetWMPing (Client *c)
+{
+    g_return_if_fail (c != NULL);
+
+    TRACE ("entering clientSendNetWMPing");
+
+    if (c->ping_timeout_id)
+    {
+        g_source_remove (c->ping_timeout_id);
+    }
+    c->ping_timeout_id = 0;
+    c->ping_time = 0;
+}
+
+void
+clientReceiveNetWMPong (ScreenInfo *screen_info, guint32 timestamp)
+{
+    Client *c;
+    int i;
+
+    g_return_if_fail (screen_info != NULL);
+    g_return_if_fail (timestamp != CurrentTime);
+
+    TRACE ("entering clientReceiveNetWMPong, timestamp %u", (unsigned int) timestamp);
+
+    for (c = screen_info->clients, i = 0; i < screen_info->client_count; c = c->next, i++)
+    {
+        if (c->ping_time == timestamp)
+        {
+            clientRemoveNetWMPing (c);
+        }
+    }
+}
+
+gboolean
+clientSendNetWMPing (Client *c, guint32 timestamp)
+{
+    ScreenInfo *screen_info;
+    DisplayInfo *display_info;
+
+    g_return_val_if_fail (c != NULL, FALSE);
+
+    TRACE ("entering clientSendNetWMPing");
+
+    if (!FLAG_TEST (c->wm_flags, WM_FLAG_PING))
+    {
+        return (FALSE);
+    }
+
+    clientRemoveNetWMPing (c);
+
+    screen_info = c->screen_info;
+    display_info = screen_info->display_info;
+
+    /* Makes sure the timestamp is meaningfull */
+    c->ping_time = myDisplayGetTime (display_info, timestamp);
+    g_return_val_if_fail (timestamp != CurrentTime, FALSE);
+
+    sendClientMessage (screen_info, c->window, NET_WM_PING, timestamp);
+    c->ping_timeout_id =
+        g_timeout_add_full (G_PRIORITY_DEFAULT,
+                            CLIENT_PING_TIMEOUT,
+                            (GtkFunction) ping_timeout_cb,
+                            (gpointer) c, NULL);
+    return (TRUE);
+}
+
+gboolean
+clientGetUserTime (Client * c)
+{
+    ScreenInfo *screen_info;
+    DisplayInfo *display_info;
+
+    g_return_val_if_fail (c != NULL, FALSE);
+    g_return_val_if_fail (c->window != None, FALSE);
+
+    screen_info = c->screen_info;
+    display_info = screen_info->display_info;
+
+    /*
+     * We can use "c->user_time_win" safely here because this will be
+     * the same as "c->window" if the app does not support the protocol
+     * NET_WM_USER_TIME_WINDOW
+     */
+
+    if (getNetWMUserTime (display_info, c->user_time_win, &c->user_time) && (c->user_time != 0))
+    {
+        myDisplaySetLastUserTime (display_info, c->user_time);
+        FLAG_SET (c->flags, CLIENT_FLAG_HAS_USER_TIME);
+
+        return TRUE;
+    }
+
+    return FALSE;
+}
+
+void
+clientAddUserTimeWin (Client * c)
+{
+    ScreenInfo *screen_info;
+    DisplayInfo *display_info;
+
+    g_return_if_fail (c != NULL);
+    g_return_if_fail (c->window != None);
+
+    screen_info = c->screen_info;
+    display_info = screen_info->display_info;
+
+    if ((c->user_time_win != None) && (c->user_time_win != c->window))
+    {
+        XSelectInput (display_info->dpy, c->user_time_win, PropertyChangeMask);
+    }
+}
+
+void
+clientRemoveUserTimeWin (Client * c)
+{
+    ScreenInfo *screen_info;
+    DisplayInfo *display_info;
+
+    g_return_if_fail (c != NULL);
+    g_return_if_fail (c->window != None);
+
+    screen_info = c->screen_info;
+    display_info = screen_info->display_info;
+
+    if ((c->user_time_win != None) && (c->user_time_win != c->window))
+    {
+        XSelectInput (display_info->dpy, c->user_time_win, NoEventMask);
+    }
+}
